# converter.py
# Copyright (C) 2022 Red Hat, Inc.
#
# Authors:
#   Akira TAGOH  <tagoh@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Module to convert a RPM spec file."""

import argparse
import shutil
import subprocess
import sys
from collections import OrderedDict
from pyrpm.spec import Spec
from typing import Any
try:
    import _debugpath  # noqa: F401
except ModuleNotFoundError:
    pass
import pyfontrpmspec.errors as err
from pyfontrpmspec import font_reader as fr
from pyfontrpmspec.messages import Message as m
from pyfontrpmspec import sources as src
from pyfontrpmspec import template
from pyfontrpmspec.package import Package


def validate_exdata(exdata):
    """Validate exdata returned by sources.extract."""
    if 'licenses' not in exdata:
        raise TypeError(m().error('No license files detected'))
    if 'fonts' not in exdata:
        raise TypeError(m().error('No fonts files detected'))
    if 'fontconfig' not in exdata:
        raise TypeError(m().error('No fontconfig files detected'))


def params(func):
    """Decorate function to initialize default parameters."""

    def wrapper(*args, **kwargs):
        kwargs.update(zip(func.__code__.co_varnames, args))
        # Add default values for optional parameters.
        'sourcedir' not in kwargs and kwargs.update({'sourcedir': '.'})
        'excludepath' not in kwargs and kwargs.update({'excludepath': []})
        'ignore_error' not in kwargs and kwargs.update({'ignore_error': []})

        return func(**kwargs)

    return wrapper


@params
def old2new(specfile: str, **kwargs: Any) -> str:
    """Convert `specfile` to new one against Packaging Guidelines.

    Currently following keyward arguments are supported:

    'specfile': str - RPM SPEC filename to convert.
    'sourcedir': str (optional) - Source directory. current directory
                                  will be used if not.
    'excludepath': list[str] (optional) - A list of exclusive paths
                                          for sources.
    'ignore_error': list[str] (optional) - A list of exception name to ignore.
    """
    kwargs['specfile'] = specfile

    if not shutil.which('rpmspec'):
        raise AttributeError(m().error('rpmspec is not installed'))
    origspec = Spec.from_file(specfile)
    ss = subprocess.run(['rpmspec', '-P', specfile], stdout=subprocess.PIPE)
    spec = Spec.from_string(ss.stdout.decode('utf-8'))
    exdata = src.extract(spec.name, spec.version, spec.sources, **kwargs)
    extra_headers = {}
    spec.epoch and extra_headers.update({'Epoch': spec.epoch})

    validate_exdata(exdata)
    if len(spec.patches) > 1:
        for p in spec.patches:
            m([': ']).info(p).warning(
                'Ignoring patch file. they have to be done manually').out()

    exdata['setup'] = '-c -T' if not exdata['archive'] else '' if not exdata[
        'root'] else '-c' if exdata['root'] == '.' else '-n {}'.format(
            exdata['root'])
    families = []
    fontconfig = []
    for k, v in OrderedDict(fr.group(exdata['fontinfo']).items()).items():
        if 'fontmap' in exdata and k in exdata['fontmap']:
            k = exdata['fontmap'][k]
        summary = None
        description = None
        for p in spec.packages:
            if Package.is_targeted_package(p.name, exdata['foundry'], k):
                (summary,
                 description) = (spec.summary, spec.description
                                 ) if p.name == spec.name else (p.summary,
                                                                p.description)
        if not summary:
            m([': ']).info(k).warning(
                ('Unable to guess the existing package name. '
                 'some information may be missing in the spec file')).out()
        if k not in exdata['fontconfig']:
            m([': ']).info(k).warning('No fontconfig file')
        info = {
            'family': k,
            'summary': summary,
            'fonts': ' '.join([vv['file'] for vv in v]),
            'exfonts': '%{nil}',
            'conf':
            len(families) + 10 if k in exdata['fontconfig'] else '%{nil}',
            'exconf': '%{nil}',
            'description': description,
        }
        families.append(info)
        if k in exdata['fontconfig']:
            fontconfig.append(exdata['fontconfig'][k].name)
    data = {
        'version':
        spec.version,
        'release':
        origspec.release,
        'url':
        spec.url,
        'extra_headers':
        '\n'.join(['{}: {}'.format(k, v) for k, v in extra_headers.items()]),
        'common_description':
        '',
        'origsource':
        origspec.sources[0],
        'source':
        spec.sources[0],
        'copy_source':
        not exdata['archive'],
        'exsources':
        exdata['sources'],
        'nsources':
        exdata['nsources'],
        'patches':
        spec.patches,
        'license':
        spec.license,
        'license_file':
        ' '.join([s.name for s in exdata['licenses']]),
        'docs':
        ' '.join([s.name
                  for s in exdata['docs']]) if 'docs' in exdata else '%{nil}',
        'foundry':
        exdata['foundry'],
        'fonts':
        families,
        'fontconfig':
        fontconfig,
        'setup':
        exdata['setup'],
        'changelog':
        origspec.changelog.rstrip(),
    }
    if len(spec.packages) == 1:
        if len(exdata['fontconfig']) > 1:
            m().error('Multiple fontconfig files detected').throw(
                err.DuplidateFileError, kwargs['ignore_error'])
        data['family'] = families[0]['family']
        if data['family'] not in exdata['fontconfig']:
            m().error('No fontconfig file for').message(data['family']).throw(
                err.DuplidateFileError, kwargs['ignore_error'])
        data['summary'] = families[0]['summary']
        data['description'] = families[0]['description']
        data['fontconfig'] = '%{nil}' if len(
            data['fontconfig']) == 0 else data['fontconfig'][0]
        data['fonts'] = families[0]['fonts']

    return template.get(len(spec.packages), data)


def main():
    """Endpoint function to convert a RPM spec file from given parameters."""
    parser = argparse.ArgumentParser(
        description='Fonts RPM spec file converter against guidelines',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--sourcedir', default='.', help='Source directory')
    parser.add_argument('-o',
                        '--output',
                        default='-',
                        type=argparse.FileType('w'),
                        help='Output file')
    parser.add_argument('--ignore-error',
                        nargs='*',
                        help='Deal with the specific error as warning')
    parser.add_argument('SPEC', help='Spec file to convert')

    args = parser.parse_args()

    templates = old2new(args.SPEC,
                        sourcedir=args.sourcedir,
                        ignore_error=args.ignore_error)
    if templates is None:
        sys.exit(1)

    args.output.write(templates['spec'])
    args.output.close()
    print('\n', flush=True, file=sys.stderr)
    if args.output.name != '<stdout>':
        r = Package.source_name(args.output.name)
        if r is None:
            m().warning('Unable to guess the spec filename').out()
        elif r + '.spec' != args.output.name:
            m().message('Proposed spec filename is').info(r + '.spec').out()

    m([': ', ' ']).warning('Note').message(
        ('You have to review the result. '
         'this doesn\'t guarantee that the generated spec file'
         ' can be necessarily built properly.')).out()


if __name__ == '__main__':
    main()



# generator.py
# Copyright (C) 2022 Red Hat, Inc.
#
# Authors:
#   Akira TAGOH  <tagoh@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Module to generate a RPM spec file."""

import argparse
import getpass
import json
import os
import pwd
import re
import sys
import textwrap
from collections import OrderedDict
from datetime import date
from babel.dates import format_date
from pathlib import Path
from typing import Any
from urllib.parse import urlparse
try:
    import _debugpath  # noqa: F401
except ModuleNotFoundError:
    pass
import pyfontrpmspec.errors as err
from pyfontrpmspec import font_reader as fr
from pyfontrpmspec.messages import Message as m
from pyfontrpmspec import sources as src
from pyfontrpmspec import template
from pyfontrpmspec.package import Package, FamilyString


def params(func):
    """Decorate function to initialize default parameters."""

    def wrapper(*args, **kwargs):
        kwargs.update(zip(func.__code__.co_varnames, args))
        # Add default values for optional parameters.
        'alias' not in kwargs and kwargs.update({'alias': 'auto'})
        'changelog' not in kwargs and kwargs.update(
            {'changelog': 'Initial import'})
        'description' not in kwargs and kwargs.update({
            'description': ('This package contains {family} which is a {alias}'
                            ' typeface of {type} font.')
        })
        'epoch' not in kwargs and kwargs.update({'epoch': None})
        'common_description' not in kwargs and kwargs.update(
            {'common_description': ''})
        'email' not in kwargs and kwargs.update(
            {'email': os.environ.get('EMAIL')})
        'excludepath' not in kwargs and kwargs.update({'excludepath': []})
        'ignore_error' not in kwargs and kwargs.update({'ignore_error': []})
        'lang' not in kwargs and kwargs.update({'lang': None})
        'license' not in kwargs and kwargs.update({'license': 'OFL-1.1'})
        'output' not in kwargs and kwargs.update({'output': '-'})
        if not (hasattr(kwargs['output'], 'write')
                and hasattr(kwargs['output'], 'close')):
            kwargs['output'] = sys.stdout if kwargs['output'] == '-' else open(
                kwargs['output'], 'w')
        if 'sources' in kwargs and not isinstance(kwargs['sources'], list):
            kwargs['sources'] = list(kwargs['sources'])
        'outputdir' not in kwargs and kwargs.update({'outputdir': '.'})
        'priority' not in kwargs and kwargs.update({'priority': 69})
        'vf_priority' not in kwargs and kwargs.update({'vf_priority': 68})
        'sourcedir' not in kwargs and kwargs.update({'sourcedir': '.'})
        'summary' not in kwargs and kwargs.update(
            {'summary': '{family}, {alias} typeface {type} font'})
        'username' not in kwargs and kwargs.update(
            {'username': pwd.getpwnam(getpass.getuser()).pw_gecos})
        'version' not in kwargs and kwargs.update({'version': None})
        'rpmautospec' not in kwargs and kwargs.update({'rpmautospec': True})
        'autorelease_opt' not in kwargs and kwargs.update(
            {'autorelease_opt': ''})
        return func(**kwargs)

    return wrapper


@params
def generate(name: str, sources: str | list[str], url: str,
             **kwargs: Any) -> dict[str, Any]:
    """Generate a spec file.

    Currently following keyword arguments are supported:

    'name': str - Archive name.
    'sources': str|list[str] - the source files.
    'url': str - URL to the project.
    'alias': str (optional) - Alias name for targeted family.
    'changelog': str (optional) - changelog entry.
    'description': str (optional) - Package description.
    'epoch': int (optional) - Epoch number.
    'common_description': str (optional) - Common package description.
                                           This is used only when generating
                                           multi packages.
    'email': str (optional) - A mail address for maintainer.
    'excludepath': list[str] (optional) - A list of exclusive paths
                                          for sources.
    'ignore_error': list[str] (optional) - A list of exception name to ignore.
    'lang': list[str] (optional) - A list of targeted language for a font
    'license': str (optional) - License name.
    'priority': int (optional) - Number of Fontconfig config priority.
    'vf_priority': int (optional) - Number of Fontconfig config priority
                                    for variable font.
    'sourcedir': str (optional) - Source directory. current directory
                                  will be used if not.
    'summary': str (optional) - Summary of package.
    'username': str (optional) - A name of package maintainer.
    'version': str (optional) - Archive version. if not specified,
                                it will be guessed from the source.

    This function returns dict with following key and values:
    'spec': str - RPM spec
    'fontconfig': FontconfigGenerator - fontconfig file to be output
    'rpmautospec': bool - True to use rpmautospec otherwise False.
    'autorelease_opt': str - Extra arguments to %autorelease.
    """
    kwargs['name'] = name
    kwargs['sources'] = sources
    kwargs['url'] = url
    kwargs['common_description'] is None and kwargs.update(
        {'common_description': ''})
    kwargs['autorelease_opt'] is None and kwargs.update(
        {'autorelease_opt': ''})
    retval = {'spec': None, 'fontconfig': []}
    extra_headers = {}

    ma = re.match(
        r'^{}-v?(((?!tar|zip)[0-9.a-zA-Z])+)\..*'.format(kwargs['name']),
        Path(src.Source(kwargs['sources'][0]).name).name)
    version = kwargs['version'] if kwargs['version'] is not None else ma.group(
        1) if ma else None
    if version is None:
        raise TypeError(m().error('Unable to guess version number'))
    v = version.split(':')
    (kwargs['epoch'],
     kwargs['version']) = v if len(v) == 2 else (kwargs['epoch'], v[0])
    kwargs['epoch'] is not None and extra_headers.update(
        {'Epoch': kwargs['epoch']})

    exdata = src.extract(**kwargs)

    if len(exdata['licenses']) == 0:
        m().error('No license files detected').throw(err.MissingFileError,
                                                     kwargs['ignore_error'])
    if len(exdata['fonts']) == 0:
        # Ignoring this isn't useful at all. don't pass kwargs['ignore_error']
        # intentionally
        m().error('No fonts files detected').throw(err.MissingFileError)

    if not exdata['archive']:
        exdata['setup'] = '-c -T'
    elif not exdata['root']:
        exdata['setup'] = ''
    elif exdata['root'] == '.':
        exdata['setup'] = '-c'
    else:
        exdata['setup'] = '-n {}'.format(exdata['root'])

    data = {}
    families = []
    fontconfig = []
    for k, v in OrderedDict(sorted(fr.group(
            exdata['fontinfo']).items())).items():
        if len(v[0]['fontinfo']['alias']) > 1:
            m([': '
               ]).info(k).warning('Multiple generic alias was detected').info(
                   v[0]['fontinfo']['alias']).out()
        if kwargs['alias'] == 'auto':
            kwargs['alias'] = v[0]['fontinfo']['alias'][0]
        info = {
            'family':
            k,
            'summary':
            kwargs['summary'].format(family=k,
                                     alias=kwargs['alias'],
                                     type=v[0]['fontinfo']['type']),
            'fonts':
            ' '.join([vv['file'] for vv in v]),
            'exfonts':
            '%{nil}',
            'conf':
            len(families) + 10,
            'exconf':
            '%{nil}',
            'description':
            kwargs['description'].format(family=k,
                                         alias=kwargs['alias'],
                                         type=v[0]['fontinfo']['type']),
        }
        families.append(info)
        c = FontconfigGenerator()
        for a in [vvv for vv in v for vvv in vv['fontinfo']['alias']]:
            c.add(a, v[0]['fontinfo']['family'], kwargs['lang'],
                  v[0]['fontinfo']['hashint'])
        c.set_fn(
            kwargs['priority']
            if not v[0]['fontinfo']['variable'] else kwargs['vf_priority'],
            str(
                FamilyString(exdata['foundry'] + ' ' + re.sub(
                    r'^{}'.format(exdata['foundry']), '', k)).normalize()) +
            '-fonts')
        retval['fontconfig'].append(c)
        fontconfig.append(c.get_fn())

    source = kwargs['sources'][0] if urlparse(
        kwargs['sources'][0], allow_fragments=True).scheme else Path(
            kwargs['sources'][0]).name
    release = '1' if not kwargs['rpmautospec'] else '%autorelease{}{}'.format(
        '' if len(kwargs['autorelease_opt']) == 0 else ' ',
        kwargs['autorelease_opt'])
    changelog = '* {} {} <{}> - {}-1\n- {}'.format(
        format_date(date.today(), "EEE MMM dd yyyy",
                    locale='en'), kwargs['username'], kwargs['email'], version,
        kwargs['changelog']) if not kwargs['rpmautospec'] else '%autochangelog'
    data = {
        'version':
        kwargs['version'],
        'release':
        release,
        'url':
        kwargs['url'],
        'extra_headers':
        '\n'.join(['{}: {}'.format(k, v) for k, v in extra_headers.items()]),
        'common_description':
        '\n'.join(textwrap.wrap(kwargs['common_description'])),
        'source':
        source,
        'copy_source':
        not exdata['archive'],
        'exsources':
        exdata['sources'],
        'nsources':
        exdata['nsources'],
        'license':
        kwargs['license'],
        'license_file':
        ' '.join([s.name for s in exdata['licenses']]),
        'docs':
        ' '.join([s.name for s in exdata['docs']])
        if len(exdata['docs']) > 0 else '%{nil}',
        'foundry':
        exdata['foundry'],
        'fonts':
        families,
        'fontconfig':
        fontconfig,
        'setup':
        exdata['setup'],
        'changelog':
        changelog,
    }
    if len(families) == 1:
        data['family'] = families[0]['family']
        data['summary'] = families[0]['summary']
        data['description'] = '\n'.join(
            textwrap.wrap(families[0]['description']))
        data['fontconfig'] = '%{nil}' if len(
            data['fontconfig']) == 0 else data['fontconfig'][0]
        data['fonts'] = families[0]['fonts']

    retval.update(template.get(len(families), data))
    return retval


class FontconfigEntry:
    """Class to hold font information."""

    def __init__(self,
                 family: str,
                 lang: str | list[str] = None,
                 hashint: bool = False):
        """Initialize a FontconfigEntry class."""
        self.family = family
        self.lang = lang
        self.hashint = hashint


class FontconfigGenerator:
    """Class to generate a fontconfig config file."""

    def __init__(self):
        """Initialize a FontconfigGenerator class."""
        self._families = {}
        self.path = None
        self._confname = ''

    def add(self,
            alias: str,
            family: str,
            lang: str | list[str] = None,
            hashint: bool = False) -> None:
        """Add the information of fonts into the object."""
        if alias not in self._families:
            self._families[alias] = []
        if not isinstance(lang, list):
            lang = [lang]
        for v in self._families[alias]:
            if v.family == family and set(
                    v.lang) == set(lang) and v.hashint == hashint:
                return
        self._families[alias].append(FontconfigEntry(family, lang, hashint))

    def set_fn(self, priority: int, fn: str) -> None:
        """Set a filename."""
        self._confname = "{:02}-{}.conf".format(priority, fn)

    def get_fn(self) -> str:
        """Get a real filename which contains a fontconfig config."""
        return self._confname

    def write(self) -> None:
        """Write a content of fontconfig config into a file."""
        if self._confname is None:
            raise TypeError(
                m().warning('filename isn\'t yet set for fontconfig.').out())

        template = ('<?xml version="1.0"?>\n'
                    '<!DOCTYPE fontconfig SYSTEM "urn:fontconfig:fonts.dtd">\n'
                    '<fontconfig>\n'
                    '{rules}'
                    '</fontconfig>\n')
        generic = ('  <match>\n'
                   '{langrule}'
                   '    <test name="family">\n'
                   '      <string>{alias}</string>\n'
                   '    </test>\n'
                   '    <edit name="family" mode="prepend">\n'
                   '      <string>{family}</string>\n'
                   '    </edit>\n'
                   '    <edit name="fonthashint" mode="append">\n'
                   '      <bool>{hashint}</bool>\n'
                   '    </edit>\n'
                   '  </match>\n')
        default = ('  <alias>\n'
                   '    <family>{family}</family>\n'
                   '    <default>\n'
                   '      <family>{alias}</family>\n'
                   '    </default>\n'
                   '  </alias>\n')
        langrule = ('    <test name="lang" compare="contains">\n'
                    '      <string>{lang}</string>\n'
                    '    </test>\n')
        rules = []
        for k, v in self._families.items():
            for vv in v:
                for ll in vv.lang:
                    if ll is None:
                        lv = ''
                    else:
                        lv = langrule.format(lang=ll)
                    s = generic.format(langrule=lv,
                                       alias=k,
                                       family=vv.family,
                                       hashint=str(vv.hashint).lower())
                    rules.append(s)
                s = default.format(alias=k, family=vv.family)
                rules.append(s)

        if self.path is None:
            raise ValueError('Set a path first.')
        with open(Path(self.path) / self._confname, 'w') as f:
            m([': ', ' ']).info(
                self._confname).message('fontconfig file was stored at').info(
                    self.path).out()
            f.write(template.format(rules=''.join(rules)))


class dotdict(dict):
    """Wrapper class to convert dict to Object."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def __load_config(config, args):
    with open(config) as f:
        confdata = json.load(f)

    args = vars(args)
    for k, v in confdata.items():
        # override missing values only.
        # have a priority to properties given by options
        if k not in args.keys():
            args[k] = v

    return dotdict(args)


def main():
    """Endpoint function to generate a RPM spec file from given parameters."""
    parser = argparse.ArgumentParser(
        description='Fonts RPM spec file generator against guidelines',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-f',
                        '--json-file',
                        help='Config file written in JSON')
    parser.add_argument('-l',
                        '--license',
                        default='OFL-1.1',
                        help='License name of this project')
    parser.add_argument('-o',
                        '--output',
                        default='-',
                        type=argparse.FileType('w'),
                        help='Output file')
    parser.add_argument('--outputdir', default='.', help='Output directory')
    parser.add_argument('--sourcedir', default='.', help='Source directory')
    parser.add_argument('-s', '--source', action='append', help='Source file')
    parser.add_argument('-c',
                        '--changelog',
                        default='Initial import',
                        help='Changelog entry')
    parser.add_argument('--email',
                        default=os.environ.get('EMAIL'),
                        help='email address to put into changelog')
    parser.add_argument('--username',
                        default=pwd.getpwnam(getpass.getuser()).pw_gecos,
                        help='Real user name to put into changelog')
    parser.add_argument('--summary',
                        default='{family}, {alias} typeface {type} font',
                        help='Summary text for package')
    parser.add_argument(
        '--description',
        default=('This package contains {family} which is a {alias}'
                 ' typeface of {type} font.'),
        help='Package description')
    parser.add_argument(
        '--common-description',
        help=('Common package description. '
              'this is only used when generating multi packages.'))
    parser.add_argument('-a',
                        '--alias',
                        default='auto',
                        help=('Set an alias name for family, '
                              'such as sans-serif, serif, monospace'))
    parser.add_argument('--lang',
                        nargs='*',
                        help='Targetted language for a font')
    parser.add_argument('--priority',
                        type=int,
                        default=69,
                        help='Number of Fontconfig config priority')
    parser.add_argument(
        '--vf-priority',
        type=int,
        default=68,
        help='Number of Fontconfig config priority for variable font')
    parser.add_argument('-e',
                        '--excludepath',
                        action='append',
                        help='Exclude path from source archives')
    parser.add_argument('--rpmautospec',
                        action=argparse.BooleanOptionalAction,
                        default=True,
                        help='Use rpmautospec.')
    parser.add_argument('--autorelease-opt',
                        help='Extra arguments to %autorelease.')
    parser.add_argument('--ignore-error',
                        nargs='*',
                        help='Deal with the specific error as warning')
    parser.add_argument('NAME', help='Package name')
    parser.add_argument('VERSION', nargs='?', help='Package version')
    parser.add_argument('URL', help='Project URL')

    args = parser.parse_args()
    if args.json_file:
        args = __load_config(args.json_file, args)
    if args.source is None or len(args.source) == 0:
        parser.print_usage()
        sys.exit(1)

    templates = generate(name=args.NAME,
                         version=args.VERSION,
                         url=args.URL,
                         license=args.license,
                         sources=args.source,
                         sourcedir=args.sourcedir,
                         changelog=args.changelog,
                         email=args.email,
                         username=args.username,
                         summary=args.summary,
                         description=args.description,
                         common_description=args.common_description,
                         alias=args.alias,
                         lang=args.lang,
                         priority=args.priority,
                         vf_priority=args.vf_priority,
                         excludepath=args.excludepath,
                         ignore_error=args.ignore_error,
                         rpmautospec=args.rpmautospec,
                         autorelease_opt=args.autorelease_opt)
    if templates is None:
        sys.exit(1)

    for f in templates['fontconfig']:
        f.path = args.outputdir
        f.write()
    args.output.write(templates['spec'])
    args.output.close()
    print('\n', flush=True, file=sys.stderr)
    if args.output.name != '<stdout>':
        r = Package.source_name(args.output.name)
        if r is None:
            m().warning('Unable to guess the spec filename').out()
        elif r + '.spec' != args.output.name:
            m().message('Proposed spec filename is').info(r + '.spec').out()

    m([': ', ' ']).warning('Note').message(
        ('You have to review the result. '
         'this doesn\'t guarantee that the generated spec file'
         ' can be necessarily built properly.')).out()


if __name__ == '__main__':
    main()



# messages.py
# Copyright (C) 2022 Red Hat, Inc.
#
# Authors:
#   Akira TAGOH  <tagoh@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Module to deal with messages."""

import re
import sys
from termcolor import colored
from typing import Self


class Message:
    """Colorize message text in a structured way."""

    def __init__(self, joiner: list[str] = [' ']):
        """Initialize Message class."""
        self._message = None
        self.__joiner = joiner
        self.__njoiner = 0

    def __add_joiner(self):
        if self._message is None:
            self._message = ''
        elif isinstance(self.__joiner, list):
            self._message += self.__joiner[self.__njoiner]
            self.__njoiner = min(self.__njoiner + 1, len(self.__joiner) - 1)

    def error(self, msg: str) -> Self:
        """Output `msg` as an error."""
        self.__add_joiner()
        self._message += colored(str(msg), 'red')
        return self

    def warning(self, msg: str) -> Self:
        """Output `msg` as a warning."""
        self.__add_joiner()
        self._message += colored(str(msg), 'yellow')
        return self

    def info(self, msg: str) -> Self:
        """Output `msg` as an information."""
        self.__add_joiner()
        self._message += colored(str(msg), 'green')
        return self

    def ignored(self) -> Self:
        """Mark this as ignored."""
        self.__add_joiner()
        self._message += colored('(ignored)', 'white')
        return self

    def message(self, msg: str) -> Self:
        """Output `msg` as a message."""
        self.__add_joiner()
        self._message += str(msg)
        return self

    def out(self) -> None:
        """Output all the strings held in this object into stderr."""
        print(self._message, flush=True, file=sys.stderr)

    def throw(self, klass, exclude: list[str] = []) -> None:
        """Raise exception."""
        if exclude is None:
            exclude = []
        for n in exclude:
            if re.match(r'{}'.format(n.lower()), klass.__name__.lower()):
                self.ignored().out()
                return
        raise klass(self._message)

    def __str__(self) -> str:
        """Convert messages to str."""
        return self._message


if __name__ == '__main__':
    Message([': ', ' ']).info('foo').warning('duplicate').out()



# Copyright (C) 2023 font-rpm-spec-generator Authors
# SPDX-License-Identifier: GPL-3.0-or-later
"""Exceptions module."""


class DuplidateFileError(TypeError):
    """Raise when duplicate files detected."""

    pass


class FileNotFoundError(AttributeError):
    """Raise when file is not available."""

    pass


class MissingFileError(ValueError):
    """Raise when there are any missing files."""

    pass



# font_reader.py
# Copyright (C) 2021-2022 Red Hat, Inc.
#
# Authors:
#   Vishal Vijayraghavan <vvijayra AT redhat DOT com>
#   Akira TAGOH  <tagoh@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Module to deal with font metadata."""

import re
from typing import Any
from fontTools.ttLib import TTFont
try:
    import _debugpath  # noqa: F401
except ModuleNotFoundError:
    pass
from pyfontrpmspec.messages import Message as m

NAME_TABLE = {
    0: 'CopyrightNotice',
    1: 'Font_Family',
    2: 'SubFamily',
    3: 'Unique_Font_Identifier',
    4: 'Full_Font_Name',
    5: 'Version',
    6: 'PostScript_Name',
    7: 'Trademark',
    8: 'Manufacturer_Name',
    9: 'Designer',
    10: 'Description',
    11: 'Vendor_URL',
    12: 'Designer_URL',
    13: 'License_Description',
    14: 'License_Info_URL',
    15: 'Reserved',
    16: 'Typographic_Family',
    17: 'Typographic_SubFamily',
    18: 'Compatible_Full',
    19: 'Sample_Text',
    20: 'PostScript_CID_Find_Fontname',
    21: 'WWS_Family_Name',
    22: 'WWS_SubFamily_Name',
    23: 'Light_Background_Pallete',
    24: 'Dark_Background_Pallete',
    25: 'Variations_PostScript_Name_Prefix'
}


def transform_foundry(id: str) -> str:
    """Transform 4-character foundry string to the human-readable string."""
    """
    4 letter characters from OS/2 table isn't hard to recognize what it is.
    particularly foundry property in macro affects the package name.
    mapping it to the human readable/recognizable name.
    """
    FOUNDARIES = {
        'ADBO': 'adobe',
        'GOOG': 'google',
        'MTY ': 'Motoya',
    }
    return FOUNDARIES[id] if id in FOUNDARIES else id


def font_meta_reader(fontfile: str, font_number: int = 0) -> dict[str, Any]:
    """Read metadata from `fontfile`."""
    meta_data = dict()
    font = TTFont(fontfile, fontNumber=font_number)
    # variable fmd denotes font meta data or fonts meta attributes
    for fmd in font['name'].names:
        if (fmd.platformID == 3
                and fmd.langID == 0x0409) or (fmd.platformID == 1
                                              and fmd.langID == 0):
            meta_data[NAME_TABLE.get(fmd.nameID, False)] = fmd.toStr()
    meta_data['foundry'] = transform_foundry(font['OS/2'].achVendID)
    meta_data['family'] = get_better_family(meta_data)
    meta_data[
        'type'] = 'OpenType' if font.sfntVersion == 'OTTO' else 'TrueType'
    fc = FontClass(fontfile, faceId=font_number)
    meta_data['alias'] = fc.get_alias_name()
    meta_data['hashint'] = True if 'prep' in font or 'cvt' in font else False
    meta_data['variable'] = True if 'fvar' in font and 'gvar' in font else False
    return meta_data


def get_better_family(meta: dict[str, Any]) -> str:
    """Get better family name from metadata."""
    if 'WWS_Family_Name' in meta:
        family = meta['WWS_Family_Name']
    elif 'Typographic_Family' in meta:
        family = meta['Typographic_Family']
    else:
        family = meta['Font_Family']
    return family


def group(families: dict[str, Any]) -> dict[str, dict[str, Any]]:
    """Restructure metadata against related family names."""
    retval = {}
    x = sorted(families.items(), key=lambda x: len(x[1]['family']))
    for k, v in x:
        found = False
        family = v['family'] if not v['variable'] else v['family'] + ' VF'
        for f in retval.keys():
            if re.fullmatch(r'{}'.format(f), family):
                retval[family].append({'fontinfo': v, 'file': k})
                found = True
        if not found:
            retval[family] = [{'fontinfo': v, 'file': k}]
    return retval


class FontClass:
    """Classified font class."""

    TYPE_SANS_SERIF = 0
    TYPE_SERIF = 1
    TYPE_MONOSPACE = 2
    TYPE_CURSIVE = 3
    TYPE_FANTASY = 4
    TYPE_EMOJI = 5
    TYPE_MATH = 6
    TYPE_END = 7

    def __init__(self, fn: str, faceId: int = -1):
        """Initialize `FontClass`."""
        self.file = fn
        self.index = faceId

    def __get_type_id(self, n):
        return 1 << n

    def __guess_class(self):
        retval = 0
        tt = TTFont(self.file, fontNumber=self.index)
        os2 = tt['OS/2']
        cls_id = (os2.sFamilyClass >> 8) & 0xff
        subcls_id = os2.sFamilyClass & 0xff
        match cls_id:
            case 0 | 6 | 11 | 13 | 14:
                pass
            case 1:
                retval |= self.__get_type_id(FontClass.TYPE_SERIF)
                if subcls_id == 8:
                    retval |= self.__get_type_id(FontClass.TYPE_CURSIVE)
            case 2 | 3:
                retval |= self.__get_type_id(FontClass.TYPE_SERIF)
                if subcls_id == 2:
                    retval |= self.__get_type_id(FontClass.TYPE_CURSIVE)
            case 4 | 5 | 7:
                retval |= self.__get_type_id(FontClass.TYPE_SERIF)
            case 8:
                retval |= self.__get_type_id(FontClass.TYPE_SANS_SERIF)
            case 9 | 12:
                retval |= self.__get_type_id(FontClass.TYPE_FANTASY)
            case 10:
                retval |= self.__get_type_id(FontClass.TYPE_CURSIVE)
            case _:
                m([':', ': ']).info(self.file).info(self.index).warning(
                    'Unknown sFamilyClass class ID: {}'.format(cls_id)).out()
        ft = os2.panose.bFamilyType
        ss = os2.panose.bSerifStyle
        pp = os2.panose.bProportion
        if ft == 3:
            retval |= self.__get_type_id(FontClass.TYPE_CURSIVE)
        elif ft >= 4 and ft <= 5:
            retval |= self.__get_type_id(FontClass.TYPE_FANTASY)
        elif ft == 2 and ((ss >= 11 and ss <= 13) or ss == 15):
            retval |= self.__get_type_id(FontClass.TYPE_SANS_SERIF)
        elif ft == 1:
            pass
        elif ft == 0 and ss == 1:
            pass
        else:
            retval |= self.__get_type_id(FontClass.TYPE_SERIF)
        if (ft == 2 and pp == 9) or (ft == 3 and pp == 3) or (
                ft == 4 and pp == 9) or (ft == 5 and pp == 3):
            retval |= self.__get_type_id(FontClass.TYPE_MONOSPACE)

        return retval

    def get_alias_name(self) -> list[str]:
        """Get alias name corresponding to the font."""
        alias = [
            'sans-serif', 'serif', 'monospace', 'cursive', 'fantasy', 'emoji',
            'math'
        ]
        id = self.__guess_class()
        retval = []
        if id == 0:
            return ''
        for i in range(FontClass.TYPE_END):
            if (id & self.__get_type_id(i)) == self.__get_type_id(i):
                retval.append(alias[i])

        return retval



# sources.py
# Copyright (C) 2022 Red Hat, Inc.
#
# Authors:
#   Akira TAGOH  <tagoh@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
import unittest
import sys
sys.path.append(str(Path(__file__).parents[1]))
from pyfontrpmspec import sources as src

class TestSource(unittest.TestCase):

    def setUp(self):
        self.license_from_source = src.Source('LICENSE.txt', '/path/to/source')
        self.doc_from_source = src.Source('README.txt', '/path/to/source')
        self.font_from_source = src.Source('foo.ttf', '/path/to/source')
        self.font_from_web = src.Source('http://example.com/foo.ttf', '/path/to/source')
        self.font_from_web_frag = src.Source('http://example.com/foo.ttf#/bar.ttf', '/path/to/source')
        self.font_from_web_query = src.Source('http://android.git.kernel.org/?p=platform/frameworks/base.git;a=blob_plain;f=data/fonts/MTLc3m.ttf', '/path/to/source')
        self.fc_from_source = src.Source('foo.conf', '/path/to/source')
        self.archive_from_source = src.Source('foo.zip', '/path/to/source')
        self.archive_from_web = src.Source('http://example.com/foo.zip', '/path/to/source')
        self.archive_from_web_frag = src.Source('http://example.com/foo#/bar.zip', '/path/to/source')

    def test_name(self):
        self.assertEqual(self.license_from_source.name, 'LICENSE.txt')
        self.assertEqual(self.doc_from_source.name, 'README.txt')
        self.assertEqual(self.font_from_source.name, 'foo.ttf')
        self.assertEqual(self.font_from_web.name, 'foo.ttf')
        self.assertEqual(self.font_from_web_frag.name, 'bar.ttf')
        self.assertEqual(self.font_from_web_query.name, 'MTLc3m.ttf')
        self.assertEqual(self.fc_from_source.name, 'foo.conf')
        self.assertEqual(self.archive_from_source.name, 'foo.zip')
        self.assertEqual(self.archive_from_web.name, 'foo.zip')
        self.assertEqual(self.archive_from_web_frag.name, 'bar.zip')

    def test_realname(self):
        self.assertEqual(self.license_from_source.realname, 'LICENSE.txt')
        self.assertEqual(self.doc_from_source.realname, 'README.txt')
        self.assertEqual(self.font_from_source.realname, 'foo.ttf')
        self.assertEqual(self.font_from_web.realname, 'http://example.com/foo.ttf')
        self.assertEqual(self.font_from_web_frag.realname, 'http://example.com/foo.ttf#/bar.ttf')
        self.assertEqual(self.font_from_web_query.realname, 'http://android.git.kernel.org/?p=platform/frameworks/base.git;a=blob_plain;f=data/fonts/MTLc3m.ttf')
        self.assertEqual(self.fc_from_source.realname, 'foo.conf')
        self.assertEqual(self.archive_from_source.realname, 'foo.zip')
        self.assertEqual(self.archive_from_web.realname, 'http://example.com/foo.zip')
        self.assertEqual(self.archive_from_web_frag.realname, 'http://example.com/foo#/bar.zip')

    def test_fullname(self):
        self.assertEqual(self.license_from_source.fullname, '/path/to/source/LICENSE.txt')
        self.assertEqual(self.doc_from_source.fullname, '/path/to/source/README.txt')
        self.assertEqual(self.font_from_source.fullname, '/path/to/source/foo.ttf')
        self.assertEqual(self.font_from_web.fullname, '/path/to/source/foo.ttf')
        self.assertEqual(self.font_from_web_frag.fullname, '/path/to/source/bar.ttf')
        self.assertEqual(self.font_from_web_query.fullname, '/path/to/source/MTLc3m.ttf')
        self.assertEqual(self.fc_from_source.fullname, '/path/to/source/foo.conf')
        self.assertEqual(self.archive_from_source.fullname, '/path/to/source/foo.zip')
        self.assertEqual(self.archive_from_web.fullname, '/path/to/source/foo.zip')
        self.assertEqual(self.archive_from_web_frag.fullname, '/path/to/source/bar.zip')

class TestFile(unittest.TestCase):

    def setUp(self):
        self.license_from_source = src.File('LICENSE.txt', '/path/to/source', True)
        self.doc_from_source = src.File('README.txt', '/path/to/source', True)
        self.doc_from_archive = src.File('./README.md', '/tmp/foo', False)
        self.doc_from_archive_sub = src.File('foo-1.0/README.md', '/tmp/foo', False)
        self.font_from_source = src.File('foo.ttf', '/path/to/source', True)
        self.font_from_web = src.File('http://example.com/baz/foo.ttf', '/path/to/source', True)
        self.font_from_web_frag = src.File('http://example.com/foo.ttf#/bar.ttf', '/path/to/source', True)
        self.font_from_web_query = src.File('http://android.git.kernel.org/?p=platform/frameworks/base.git;a=blob_plain;f=data/fonts/MTLc3m.ttf', '/path/to/source', True)
        self.font_from_archive = src.File('ttf/foo.ttf', '/tmp/foo', False)
        self.font_from_archive_sub = src.File('foo-1.0/ttf/foo.ttf', '/tmp/foo', False)
        self.fc_from_source = src.File('foo.conf', '/path/to/source', True)

    def test_name(self):
        self.assertEqual(self.license_from_source.name, 'LICENSE.txt')
        self.assertEqual(self.doc_from_source.name, 'README.txt')
        self.assertEqual(self.doc_from_archive.name, 'README.md')
        self.assertEqual(self.doc_from_archive_sub.name, 'README.md')
        self.assertEqual(self.font_from_source.name, 'foo.ttf')
        self.assertEqual(self.font_from_web.name, 'foo.ttf')
        self.assertEqual(self.font_from_web_frag.name, 'bar.ttf')
        self.assertEqual(self.font_from_web_query.name, 'MTLc3m.ttf')
        self.assertEqual(self.font_from_archive.name, 'foo.ttf')
        self.assertEqual(self.font_from_archive_sub.name, 'ttf/foo.ttf')
        self.assertEqual(self.fc_from_source.name, 'foo.conf')

    def test_realname(self):
        self.assertEqual(self.license_from_source.realname, 'LICENSE.txt')
        self.assertEqual(self.doc_from_source.realname, 'README.txt')
        self.assertEqual(self.doc_from_archive.realname, './README.md')
        self.assertEqual(self.doc_from_archive_sub.realname, 'foo-1.0/README.md')
        self.assertEqual(self.font_from_source.realname, 'foo.ttf')
        self.assertEqual(self.font_from_web.realname, 'http://example.com/baz/foo.ttf')
        self.assertEqual(self.font_from_web_frag.realname, 'http://example.com/foo.ttf#/bar.ttf')
        self.assertEqual(self.font_from_web_query.realname, 'http://android.git.kernel.org/?p=platform/frameworks/base.git;a=blob_plain;f=data/fonts/MTLc3m.ttf')
        self.assertEqual(self.font_from_archive.realname, 'ttf/foo.ttf')
        self.assertEqual(self.font_from_archive_sub.realname, 'foo-1.0/ttf/foo.ttf')
        self.assertEqual(self.fc_from_source.realname, 'foo.conf')

    def test_fullname(self):
        self.assertEqual(self.license_from_source.fullname, '/path/to/source/LICENSE.txt')
        self.assertEqual(self.doc_from_source.fullname, '/path/to/source/README.txt')
        self.assertEqual(self.doc_from_archive.fullname, '/tmp/foo/README.md')
        self.assertEqual(self.doc_from_archive_sub.fullname, '/tmp/foo/foo-1.0/README.md')
        self.assertEqual(self.font_from_source.fullname, '/path/to/source/foo.ttf')
        self.assertEqual(self.font_from_web.fullname, '/path/to/source/foo.ttf')
        self.assertEqual(self.font_from_web_frag.fullname, '/path/to/source/bar.ttf')
        self.assertEqual(self.font_from_web_query.fullname, '/path/to/source/MTLc3m.ttf')
        self.assertEqual(self.font_from_archive.fullname, '/tmp/foo/ttf/foo.ttf')
        self.assertEqual(self.font_from_archive_sub.fullname, '/tmp/foo/foo-1.0/ttf/foo.ttf')
        self.assertEqual(self.fc_from_source.fullname, '/path/to/source/foo.conf')

    def test_prefix(self):
        self.assertEqual(self.license_from_source.prefix, '/path/to/source')
        self.assertEqual(self.doc_from_source.prefix, '/path/to/source')
        self.assertEqual(self.doc_from_archive.prefix, '/tmp/foo')
        self.assertEqual(self.doc_from_archive_sub.prefix, '/tmp/foo')
        self.assertEqual(self.font_from_source.prefix, '/path/to/source')
        self.assertEqual(self.font_from_web.prefix, '/path/to/source')
        self.assertEqual(self.font_from_web_frag.prefix, '/path/to/source')
        self.assertEqual(self.font_from_web_query.prefix, '/path/to/source')
        self.assertEqual(self.font_from_archive.prefix, '/tmp/foo')
        self.assertEqual(self.font_from_archive_sub.prefix, '/tmp/foo')
        self.assertEqual(self.fc_from_source.prefix, '/path/to/source')

    def test_is_license(self):
        self.assertEqual(self.license_from_source.is_license(), True)
        self.assertEqual(self.doc_from_source.is_license(), False)
        self.assertEqual(self.doc_from_archive.is_license(), False)
        self.assertEqual(self.doc_from_archive_sub.is_license(), False)
        self.assertEqual(self.font_from_source.is_license(), False)
        self.assertEqual(self.font_from_web.is_license(), False)
        self.assertEqual(self.font_from_web_frag.is_license(), False)
        self.assertEqual(self.font_from_web_query.is_license(), False)
        self.assertEqual(self.font_from_archive.is_license(), False)
        self.assertEqual(self.font_from_archive_sub.is_license(), False)
        self.assertEqual(self.fc_from_source.is_license(), False)

    def test_is_doc(self):
        self.assertEqual(self.license_from_source.is_doc(), True)
        self.assertEqual(self.doc_from_source.is_doc(), True)
        self.assertEqual(self.doc_from_archive.is_doc(), True)
        self.assertEqual(self.doc_from_archive_sub.is_doc(), True)
        self.assertEqual(self.font_from_source.is_doc(), False)
        self.assertEqual(self.font_from_web.is_doc(), False)
        self.assertEqual(self.font_from_web_frag.is_doc(), False)
        self.assertEqual(self.font_from_web_query.is_doc(), False)
        self.assertEqual(self.font_from_archive.is_doc(), False)
        self.assertEqual(self.font_from_archive_sub.is_doc(), False)
        self.assertEqual(self.fc_from_source.is_doc(), False)

    def test_is_font(self):
        self.assertEqual(self.license_from_source.is_font(), False)
        self.assertEqual(self.doc_from_source.is_font(), False)
        self.assertEqual(self.doc_from_archive.is_font(), False)
        self.assertEqual(self.doc_from_archive_sub.is_font(), False)
        self.assertEqual(self.font_from_source.is_font(), True)
        self.assertEqual(self.font_from_web.is_font(), True)
        self.assertEqual(self.font_from_web_frag.is_font(), True)
        self.assertEqual(self.font_from_web_query.is_font(), True)
        self.assertEqual(self.font_from_archive.is_font(), True)
        self.assertEqual(self.font_from_archive_sub.is_font(), True)
        self.assertEqual(self.fc_from_source.is_font(), False)

    def test_is_fontconfig(self):
        self.assertEqual(self.license_from_source.is_fontconfig(), False)
        self.assertEqual(self.doc_from_source.is_fontconfig(), False)
        self.assertEqual(self.doc_from_archive.is_fontconfig(), False)
        self.assertEqual(self.doc_from_archive_sub.is_fontconfig(), False)
        self.assertEqual(self.font_from_source.is_fontconfig(), False)
        self.assertEqual(self.font_from_web.is_fontconfig(), False)
        self.assertEqual(self.font_from_web_frag.is_fontconfig(), False)
        self.assertEqual(self.font_from_web_query.is_fontconfig(), False)
        self.assertEqual(self.font_from_archive.is_fontconfig(), False)
        self.assertEqual(self.font_from_archive_sub.is_fontconfig(), False)
        self.assertEqual(self.fc_from_source.is_fontconfig(), True)

    def test_is_source(self):
        self.assertEqual(self.license_from_source.is_source(), True)
        self.assertEqual(self.doc_from_source.is_source(), True)
        self.assertEqual(self.doc_from_archive.is_source(), False)
        self.assertEqual(self.doc_from_archive_sub.is_source(), False)
        self.assertEqual(self.font_from_source.is_source(), True)
        self.assertEqual(self.font_from_web.is_source(), True)
        self.assertEqual(self.font_from_web_frag.is_source(), True)
        self.assertEqual(self.font_from_web_query.is_source(), True)
        self.assertEqual(self.font_from_archive.is_source(), False)
        self.assertEqual(self.font_from_archive_sub.is_source(), False)
        self.assertEqual(self.fc_from_source.is_source(), True)

if __name__ == '__main__':
    unittest.main()
